/*
 * '개발자 강대리' 블로그의 "한글 오토마타 만들기" 의 소스 코드를 
 *  자바로 변경/수정하였습니다. (http://codepedia.tistory.com/42)
 *  흔쾌히 변경 허락해주신 인형오리님께 감사 드립니다.
 *  
 *  어플명 : cqIME Qwerty Ver.
 *  버  전 : 0.60.0
 *  작성자 : cq (undercover.cq@gmail.com)
 *  작성일 : 2010.03.26
 *  
 *  수정/배포 시 노키아까페 및 인형오리님의 출처를 명시해 주세요. ^^
 *  감사합니다. 
 *  
 */

package net.cq.koreanIME;

public class AutomataKR {  
	  
    private char ingWord;				// 작성중 글자  
    private StringBuffer completeText;	// 완성 문자열
    private char lastInput;
    
    // 단어조합상태    
    private static final int HS_FIRST 			= 0;	// 초성  
    private static final int HS_FIRST_V 		= 1;    // 자음 + 자음   
    private static final int HS_FIRST_C 		= 2;    // 모음 + 모음  
    private static final int HS_MIDDLE_STATE 	= 3;    // 초성 + 모음 + 모음  
    private static final int HS_END 			= 4;    // 초성 + 중성 + 종성  
    private static final int HS_END_STATE 		= 5;    // 초성 + 중성 + 자음 + 자음  
    private static final int HS_END_EXCEPTION 	= 6;    // 초성 + 중성 + 종성(곁자음)  
    
    private static final int BASE_CODE = 0xac00;       // 기초음성(가)  
    private static final int LIMIT_MIN = 0xac00;       // 음성범위 MIN(가)  
    private static final int LIMIT_MAX = 0xd7a3;       // 음성범위 MAX  
    
    // SOUND_TABLE의 인덱스를 찾기 위해 사용함
	private static final String stringIndex = "rRseEfaqQtTdwWczxvgkoiOjpuPh###yn###bm#l";

    // 음성 테이블  
    private static final char[] SOUND_TABLE =  
    {  
        /* 초성 19자 0 ~ 18 */  
        ('ㄱ'), ('ㄲ'), ('ㄴ'), ('ㄷ'), ('ㄸ'),  
        ('ㄹ'), ('ㅁ'), ('ㅂ'), ('ㅃ'), ('ㅅ'), 
        ('ㅆ'), ('ㅇ'), ('ㅈ'), ('ㅉ'), ('ㅊ'),  
        ('ㅋ'), ('ㅌ'), ('ㅍ'), ('ㅎ'),  
        /* 중 성 21자 19 ~ 39 */  
        ('ㅏ'), ('ㅐ'), ('ㅑ'), ('ㅒ'), ('ㅓ'),  
        ('ㅔ'), ('ㅕ'), ('ㅖ'), ('ㅗ'), ('ㅘ'),  
        ('ㅙ'), ('ㅚ'), ('ㅛ'), ('ㅜ'), ('ㅝ'),  
        ('ㅞ'), ('ㅟ'), ('ㅠ'), ('ㅡ'), ('ㅢ'),  
        ('ㅣ'),  
        /* 종 성 28자 40 ~ 67 */  
        (' '), ('ㄱ'), ('ㄲ'), ('ㄳ'), ('ㄴ'),  
        ('ㄵ'), ('ㄶ'), ('ㄷ'), ('ㄹ'), ('ㄺ'),   
        ('ㄻ'), ('ㄼ'), ('ㄽ'), ('ㄾ'), ('ㄿ'),  
        ('ㅀ'), ('ㅁ'), ('ㅂ'), ('ㅄ'), ('ㅅ'),  
        ('ㅆ'), ('ㅇ'), ('ㅈ'), ('ㅊ'), ('ㅋ'),  
        ('ㅌ'), ('ㅍ'), ('ㅎ')
    };  
    
    /*//사용하지 않음
    // 초성 합성 테이블  
    private int[][] MIXED_CHO_CONSON =  // [14][3]
    {  
        { 0, 0,15}, // ㄱ,ㄱ,ㅋ  
        {15, 0, 1}, // ㅋ,ㄱ,ㄲ  
        { 1, 0, 0}, // ㄲ,ㄱ,ㄱ  
      
        { 3, 3,16}, // ㄷ,ㄷ,ㅌ  
        {16, 3, 4}, // ㅌ,ㄷ,ㄸ  
        { 4, 3, 3}, // ㄸ,ㄷ,ㄷ  
      
        { 7, 7,17}, // ㅂ,ㅂ,ㅍ  
        {17, 7, 8}, // ㅍ,ㅂ,ㅃ  
        { 8, 7, 7}, // ㅃ,ㅂ,ㅂ  
      
        { 9, 9,10}, // ㅅ,ㅅ,ㅆ  
        {10, 9, 9}, // ㅆ,ㅅ,ㅅ  
      
        {12,12,14}, // ㅈ,ㅈ,ㅊ  
        {14,12,13}, // ㅊ,ㅈ,ㅉ  
        {13,12,12}  // ㅉ,ㅈ,ㅈ  
    };  
    */  
    
    // 초성,중성 모음 합성 테이블  
    private static final int[][] MIXED_VOWEL =   // [21][3] --> [13][3]
    {  
        //{19,19,21}, // ㅏ,ㅏ,ㅑ  
        //{21,19,19}, // ㅑ,ㅏ,ㅏ  
      
        {19,39,20}, // ㅏ,ㅣ,ㅐ  
        {21,39,22}, // ㅑ,ㅣ,ㅒ  
      
        //{23,23,25}, // ㅓ,ㅓ,ㅕ  
        //{25,23,23}, // ㅕ,ㅓ,ㅓ  
      
        {23,39,24}, // ㅓ,ㅣ,ㅔ  
        {25,39,26}, // ㅕ,ㅣ,ㅖ  
      
        //{27,27,31}, // ㅗ,ㅗ,ㅛ  
        //{31,27,27}, // ㅛ,ㅗ,ㅗ  
      
        {27,19,28}, // ㅗ,ㅏ,ㅘ  
        {28,39,29}, // ㅘ,ㅣ,ㅙ 
      
        {27,39,30}, // ㅗ,ㅣ,ㅚ 
        
        // 새로 추가함
        {27,20,29}, // ㅗ,ㅐ,ㅙ
        {32,24,34}, // ㅜ,ㅔ,ㅞ
      
        //{32,32,36}, // ㅜ,ㅜ,ㅠ  
        //{36,32,32}, // ㅠ,ㅜ,ㅜ  
      
        {32,23,33}, // ㅜ,ㅓ,ㅝ  
        {33,39,34}, // ㅝ,ㅣ,ㅞ  
      
        {32,39,35}, // ㅜ,ㅣ,ㅟ  
      
        //{39,39,37}, // ㅣ,ㅣ,ㅡ  
        {37,39,38}//, // ㅡ,ㅣ,ㅢ  
        //{38,39,39}  // ㅢ,ㅣ,ㅣ  
    };  
      
    // 종성 합성 테이블  
    private static final int[][] MIXED_JONG_CONSON =   //[22][3]  --> [10][3]
    {  
        //{41,41,64}, // ㄱ,ㄱ,ㅋ  
        //{64,41,42}, // ㅋ,ㄱ,ㄲ  
        //{42,41,41}, // ㄲ,ㄱ,ㄱ  
      
        {41,59,43}, // ㄱ,ㅅ,ㄳ  
      
        {44,62,45}, // ㄴ,ㅈ,ㄵ  
        {44,67,46}, // ㄴ,ㅎ,ㄶ  
      
        //{47,47,65}, // ㄷ,ㄷ,ㅌ  
        //{65,47,47}, // ㅌ,ㄷ,ㄷ  
      
        {48,41,49}, // ㄹ,ㄱ,ㄺ  
        {48,56,50}, // ㄹ,ㅁ,ㄻ  
      
        {48,57,51}, // ㄹ,ㅂ,ㄼ  
        //{51,57,54}, // ㄼ,ㅂ,ㄿ  
      
        {48,59,52}, // ㄹ,ㅅ,ㄽ
        {48,65,53}, // ㄹ,ㅌ,ㄾ	
        {48,67,55}, // ㄹ,ㅎ,ㅀ  
      
        //{57,57,66}, // ㅂ,ㅂ,ㅍ  
        //{66,57,57}, // ㅍ,ㅂ,ㅂ  
      
        {57,59,58}, // ㅂ,ㅅ,ㅄ  
      
        //{59,59,60}, // ㅅ,ㅅ,ㅆ  
        //{60,59,59}, // ㅆ,ㅅ,ㅅ  
      
        //{62,62,63}, // ㅈ,ㅈ,ㅊ  
        //{63,62,62}  // ㅊ,ㅈ,ㅈ  
    };  
      
    // 종성 분해 테이블  
    private static final int[][] DIVIDE_JONG_CONSON =   // [13][3] --> [11][3]
    {  
        //{41,41,42}, // ㄱ,ㄱ,ㄲ  
        {41,59,43}, // ㄱ,ㅅ,ㄳ  
        {44,62,45}, // ㄴ,ㅈ,ㄵ  
        {44,67,46}, // ㄴ,ㅎ,ㄶ  
        {48,41,49}, // ㄹ,ㄱ,ㄺ  
        {48,56,50}, // ㄹ,ㅁ,ㄻ  
        {48,57,51}, // ㄹ,ㅂ,ㄼ  
        {51,66,54}, // ㄹ,ㅍ,ㄿ  
        {48,59,52}, // ㄹ,ㅅ,ㄽ  
        {48,65,53}, // ㄹ,ㅌ,ㄾ      
        {48,67,55}, // ㄹ,ㅎ,ㅀ  
        {57,59,58} // ㅂ,ㅅ,ㅄ  
        //{59,59,60}  // ㅅ,ㅅ,ㅆ  
    };  
    

    private int wordStatus;      // 단어조합상태  
    private char completeWord;   // 완성글자
    private int[] curWordPhonemes = {0, 0, 0, 0, 0};  // 음소[초,중,종,곁자음1,곁자음2] 
    private int[] prevWordPhonemes = {0, 0, 0, 0, 0}; // completeText의 마지막 음절 저장   
    
    public AutomataKR()  
    {  
    	completeText    = new StringBuffer(); 
        clear();  
    }  
      
    // 초기화  
    private void clear()  
    {  
        wordStatus      = HS_FIRST;  
        completeText.setLength(0);  
        ingWord         = 0;  
        lastInput 		= 0;
        completeWord    = 0; 

        phonemesClear();
        prevWordClear();
    }
    
    private void phonemesClear() {
    	for (int i = 0; i < 5; i++)
    		curWordPhonemes[i] = 0;
    }
    
    private void prevWordClear() {
    	for (int i = 0; i < 5; i++)
        	prevWordPhonemes[i] = 0;
    } 
    
    public void clearInputText() {
		ingWord = '0';
		completeText.setLength(0);
		clear();
	}
	
	// 입력 텍스트
	public String getTextResult() {
		return completeText.toString() + String.valueOf(ingWord);
	}
	
	// 입력 텍스트 길이
	public int getTextLength() {
		if (ingWord != 0) {
			return completeText.length() + 1;
		}
		else {
			return completeText.length();
		}
	}
	
	public char getLastInput() {
		return lastInput;
	}
    
    private void copyFromPhonemez() {
    	prevWordPhonemes[0] = curWordPhonemes[0];
    	prevWordPhonemes[1] = curWordPhonemes[1];
    	prevWordPhonemes[2] = curWordPhonemes[2];
    	prevWordPhonemes[3] = curWordPhonemes[3];
    	prevWordPhonemes[4] = curWordPhonemes[4];
    }
    
    private void copyToPhonemez() {
    	curWordPhonemes[0] = prevWordPhonemes[0];
    	curWordPhonemes[1] = prevWordPhonemes[1];
    	curWordPhonemes[2] = prevWordPhonemes[2];
    	curWordPhonemes[3] = prevWordPhonemes[3];
    	curWordPhonemes[4] = prevWordPhonemes[4];
    }
    
    
    // backspace 입력
    public void insertBackspace() {
    	
    	if (ingWord == 0)  
        {  
        	int index = completeText.length();
        	
            if (index > 0) {
                completeText.setLength(index-1);  
            }
        }  
        else  
            wordStatus = downGradeIngWordStatus(ingWord);
    }
    
    // 영문자 및 특수문자 입력
    public void insertSpecialKey(String ch, boolean check){
      
        // 완성 문자열 만들기  
        if(ingWord != 0) {   
            completeText.append(ingWord);
        }
        
        ingWord = 0;
        completeWord = 0;
        
        completeText.append(ch);
        wordStatus = HS_FIRST;
        
        return;
    }
    
    // 이지한글
    // 한글 입력 및 조합 (정해진 int 코드값을 입력 받아 한글조합)  
    public void insertEZHangul(int nKeyCode, boolean insert)
    {  
        completeWord = 0;
        lastInput = (char)nKeyCode;
        
        // 키코드 변환
        nKeyCode = stringIndex.indexOf((char)nKeyCode);
        
        // *, # 버튼 처리
        if (!insert) {
        	wordStatus = downGradeIngWordStatus(ingWord);
        }
        
        switch(wordStatus)  
        {
        case HS_FIRST:  
            // 초성
        	//System.out.println("HS_FIRST");
        	
    		if (nKeyCode > 18)	{ //모음
    			
    			if (completeText.length() > 0) {	// 앞 글자 있음
    				// 앞글자 분해
    				prevWordClear();
    				decomposeWord(completeText.charAt(completeText.length()-1));
    				
    				if (prevWordPhonemes[2] > 40) {	// 받침이 있다면
    					// 곁자음이라면 분해
    					if (decomposeConsonant2()) {	// 겹자음
    						
    						// 뒤의 것 빼고 합성
    						copyToPhonemez();
    						char modChar = combineHangle(3);
    						
    						// 뒤 받침과 모음 합성
    						curWordPhonemes[0]  = toInitial(prevWordPhonemes[4]);
    						curWordPhonemes[1]  = nKeyCode;  
    						wordStatus       = HS_MIDDLE_STATE;
    					
    						// 완성 문자열의 마지막 글자로 대체
    						completeText.setLength(completeText.length()-1);
    						completeText.append(modChar);
    						
    						lastInput = 0;
    						break;
    					}
    				}
    				else if (prevWordPhonemes[2] == 40) {	// 받침 없음 -> 모음합성여부 확인
    					// 모음 합성여부 확인
    					int checkVowel = mixVowel(prevWordPhonemes[1], nKeyCode);
    					if (checkVowel != 1000) {	// 합성 가능
    						
    						curWordPhonemes[0] = prevWordPhonemes[0];
    						curWordPhonemes[1] = checkVowel;
    						curWordPhonemes[2] = 40;	// 받침 없음
    						curWordPhonemes[3] = 0;
    						curWordPhonemes[4] = 0;
    						
    						completeText.setLength(completeText.length()-1);
    						wordStatus       = HS_MIDDLE_STATE;
    						lastInput = 0;
    						
    						break;
    					}
    				}
    				else {
    					System.out.println("=== Not reached 01===");
    				}
    			}
    		
    		}
    		else {	// 자음
    			if (completeText.length() > 0) {	// 앞 글자 있음
    				// 앞글자 분해
    				prevWordClear();
    				decomposeWord(completeText.charAt(completeText.length()-1));
    				
    				// 앞 글자 종성과 합성 가능한 경우
    				if (prevWordPhonemes[2] != 0) {	// 받침이 있다면
    					// 종성 변경/합성
    					int jongCode = toFinal(nKeyCode);
    					if (mixFinal2(jongCode)) {   
    						copyToPhonemez();

    						completeText.setLength(completeText.length()-1);
    						wordStatus       = HS_END_EXCEPTION;
    						combineIngWord(wordStatus);
    						return;
    					} 
    				}
    			}			
    		}
        	
        	curWordPhonemes[0]  = nKeyCode;  
			ingWord          = SOUND_TABLE[curWordPhonemes[0]];  
			wordStatus       = nKeyCode > 18 ? HS_FIRST_C : HS_FIRST_V;  
			break;
        case HS_FIRST_C:  
            // 모음 + 모음  
            //
        	//System.out.println("HS_FIRST_C");
            
            if(nKeyCode > 18)    // 모음  
            {  
            	int checkVowel = mixVowel(curWordPhonemes[0], nKeyCode);
            	
                if(checkVowel == 1000)  
                {  
                    completeWord = SOUND_TABLE[curWordPhonemes[0]];  
                    curWordPhonemes[0] = nKeyCode;  
                }
                else {
                	curWordPhonemes[0] = checkVowel;
                	lastInput = 0;
                }
            }  
            else                // 자음  
            {  
                completeWord  = SOUND_TABLE[curWordPhonemes[0]];  
                curWordPhonemes[0]  = nKeyCode;  
                wordStatus       = HS_FIRST_V;  
            }  
            break;  
      
        case HS_FIRST_V:  
            // 자음 + 자음  
            //  
        	//System.out.println("HS_FIRST_V");
        	
            if(nKeyCode > 18)    // 모음  
            {  
                curWordPhonemes[1]  = nKeyCode;  
                wordStatus       = HS_MIDDLE_STATE;  
            }  
            else                // 자음  
            {  
                //if(!mixInitial(nKeyCode))		// 자음 합성은 사용하지 않음 
                {  
                    completeWord  = SOUND_TABLE[curWordPhonemes[0]];  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;  
                }  
            }  
            break;  
      
        case HS_MIDDLE_STATE:  
            // 초성 + 모음 + 모음  
            //  
        	//System.out.println("HS_MIDDLE_STATE");
        	
        	// 받침 초기화
        	curWordPhonemes[2] = 40;
        	
            if(nKeyCode > 18)  
            {  
            	int checkVowel = mixVowel(curWordPhonemes[1], nKeyCode);
            	
                if(checkVowel == 1000)  
                {    
                    completeWord  = combineHangle(1);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_C;  
                }
                else {
                	curWordPhonemes[1] = checkVowel;
                	lastInput = 0;
                }
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
      
                if(jungCode > 0)  
                {  
                    curWordPhonemes[2]  = jungCode;  
                    wordStatus       = HS_END_STATE;  
                }  
                else  
                {  
                    completeWord  = combineHangle(1);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;
                }  
            }  
            break;  
      
        case HS_END:  
            // 초성 + 중성 + 종성  
            //  
        	//System.out.println("HS_END");
        	
            if(nKeyCode > 18)    
            {  
                completeWord  = combineHangle(1);  
                curWordPhonemes[0]  = nKeyCode;  
                wordStatus       = HS_FIRST;  
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
                if(jungCode > 0)  
                {  
                    curWordPhonemes[2]  = jungCode;  
                    wordStatus       = HS_END_STATE;  
                }  
                else  
                {  
                    completeWord  = combineHangle(1);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V; 
                }    
            }  
            break;  
      
        case HS_END_STATE:  
            // 초성 + 중성 + 자음(종) +  자음(종)  
            //  
        	//System.out.println("HS_END_STATE");
        	
            if(nKeyCode > 18)  
            {  
                completeWord = combineHangle(1);  
      
                curWordPhonemes[0]  = toInitial(curWordPhonemes[2]);  
                curWordPhonemes[1]  = nKeyCode;  
                wordStatus       = HS_MIDDLE_STATE;  
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
                if(jungCode > 0)  
                {
                    wordStatus = HS_END_EXCEPTION;  
      
                    if(!mixFinal(jungCode))  
                    {  
                        completeWord  = combineHangle(2);  
                        curWordPhonemes[0]  = nKeyCode;  
                        wordStatus       = HS_FIRST_V;
                    }  
                }  
                else  
                {  
                    completeWord  = combineHangle(2);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;  
                }    
            }  
            break;  
      
        case HS_END_EXCEPTION:  
            // 초성 + 중성 + 종성(곁자음)  
            //  
        	//System.out.println("HS_END_EXCEPTION");
        	
            if(nKeyCode > 18)    
            {  
                decomposeConsonant();  
                curWordPhonemes[1]  = nKeyCode;  
                wordStatus       = HS_MIDDLE_STATE;  
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
                if(jungCode > 0)  
                {  
                    wordStatus = HS_END_EXCEPTION;  
      
                    if(!mixFinal(jungCode))  
                    {  
                        completeWord  = combineHangle(2);  
                        curWordPhonemes[0]  = nKeyCode;  
                        wordStatus       = HS_FIRST_V;  
                    }  
                }  
                else  
                {  
                    completeWord  = combineHangle(2);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;   
                }  
            }  
            break;  
        }  
        
        copyFromPhonemez();
      
        // 현재 보이는 글자상태  
        combineIngWord(wordStatus);  
      
        // 완성 문자열 만들기  
        if(completeWord != 0) {   
            completeText.append(completeWord);
        }
      
        return;  
    }  
    
    
    // Qwerty  
    // 키코드 입력 및 조합 (정해진 int 코드값을 입력 받아 한글조합)
    public void insertQwertyHangul(int nKeyCode, boolean insert)
    {  
        completeWord = 0;
        
        // 키코드 변환
        nKeyCode = stringIndex.indexOf((char)nKeyCode);
        System.out.print(nKeyCode + ", ");
        
        switch(wordStatus)  
        {
        case HS_FIRST:  
            // 초성  
            //
        	System.out.println("HS_FIRST");      	
        	
        	curWordPhonemes[0]  = nKeyCode;  
			ingWord          = SOUND_TABLE[curWordPhonemes[0]];  
			wordStatus       = nKeyCode > 18 ? HS_FIRST_C : HS_FIRST_V;  
			break;
        case HS_FIRST_C:  
            // 모음 + 모음  
            //
        	System.out.println("HS_FIRST_C");
            
            if(nKeyCode > 18)    // 모음  
            {  
            	int checkVowel = mixVowel(curWordPhonemes[0], nKeyCode);
            	
                if(checkVowel == 1000)  
                {  
                    completeWord = SOUND_TABLE[curWordPhonemes[0]];  
                    curWordPhonemes[0] = nKeyCode;  
                }
                else {
                	curWordPhonemes[0] = checkVowel;
                }
            }  
            else                // 자음  
            {  
                completeWord  = SOUND_TABLE[curWordPhonemes[0]];  
                curWordPhonemes[0]  = nKeyCode;  
                wordStatus       = HS_FIRST_V;  
            }  
            break;  
      
        case HS_FIRST_V:  
            // 자음 + 자음  
            //  
        	System.out.println("HS_FIRST_V");
        	
            if(nKeyCode > 18)    // 모음  
            {  
                curWordPhonemes[1]  = nKeyCode;  
                wordStatus       = HS_MIDDLE_STATE;  
            }  
            else                // 자음  
            {  
                //if(!mixInitial(nKeyCode))		// 자음 합성은 사용하지 않음 
                {  
                    completeWord  = SOUND_TABLE[curWordPhonemes[0]];  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;  
                }  
            }  
            break;  
      
        case HS_MIDDLE_STATE:  
            // 초성 + 모음 + 모음  
            //  
        	System.out.println("HS_MIDDLE_STATE");
        	
            if(nKeyCode > 18)  
            {  
            	int checkVowel = mixVowel(curWordPhonemes[1], nKeyCode);
            	
                if(checkVowel == 1000)  
                {    
                    completeWord  = combineHangle(1);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_C;  
                }
                else {
                	curWordPhonemes[1] = checkVowel;
                }
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
      
                if(jungCode > 0)  
                {  
                    curWordPhonemes[2]  = jungCode;  
                    wordStatus       = HS_END_STATE;  
                }  
                else  
                {  
                    completeWord  = combineHangle(1);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;
                }  
            }  
            break;  
      
        case HS_END:  
            // 초성 + 중성 + 종성  
            //  
        	System.out.println("HS_END");
        	
            if(nKeyCode > 18)    
            {  
                completeWord  = combineHangle(1);  
                curWordPhonemes[0]  = nKeyCode;  
                wordStatus       = HS_FIRST;  
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
                if(jungCode > 0)  
                {  
                    curWordPhonemes[2]  = jungCode;  
                    wordStatus       = HS_END_STATE;  
                }  
                else  
                {  
                    completeWord  = combineHangle(1);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;
                }    
            }  
            break;  
      
        case HS_END_STATE:  
            // 초성 + 중성 + 자음(종) +  자음(종)  
            //  
        	System.out.println("HS_END_STATE");
        	
            if(nKeyCode > 18)  
            {  
                completeWord = combineHangle(1);  
      
                curWordPhonemes[0]  = toInitial(curWordPhonemes[2]);  
                curWordPhonemes[1]  = nKeyCode;  
                wordStatus       = HS_MIDDLE_STATE;  
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
                if(jungCode > 0)  
                {
                    wordStatus = HS_END_EXCEPTION;  
      
                    if(!mixFinal(jungCode))  
                    {  
                        completeWord  = combineHangle(2);  
                        curWordPhonemes[0]  = nKeyCode;  
                        wordStatus       = HS_FIRST_V;
                    }  
                }  
                else  
                {  
                    completeWord  = combineHangle(2);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;  
                }    
            }  
            break;  
      
        case HS_END_EXCEPTION:  
            // 초성 + 중성 + 종성(곁자음)  
            //  
        	System.out.println("HS_END_EXCEPTION");
        	
            if(nKeyCode > 18)    
            {  
                decomposeConsonant();  
                curWordPhonemes[1]  = nKeyCode;  
                wordStatus       = HS_MIDDLE_STATE;  
            }  
            else  
            {  
                int jungCode = toFinal(nKeyCode);  
                if(jungCode > 0)  
                {  
                    wordStatus = HS_END_EXCEPTION;  
      
                    if(!mixFinal(jungCode))  
                    {  
                        completeWord  = combineHangle(2);  
                        curWordPhonemes[0]  = nKeyCode;  
                        wordStatus       = HS_FIRST_V;  
                    }  
                }  
                else  
                {  
                    completeWord  = combineHangle(2);  
                    curWordPhonemes[0]  = nKeyCode;  
                    wordStatus       = HS_FIRST_V;   
                }  
            }  
            break;  
        }  
        
        // 현재 보이는 글자상태  
        combineIngWord(wordStatus);  
      
        // 완성 문자열 만들기  
        if(completeWord != 0) {   
            completeText.append(completeWord);
        }
      
        return;  
    }

    
    // 초성으로 변환  
    private int toInitial(int nKeyCode)  
    {  
        switch(nKeyCode)  
        {  
        case 41: return 0;  // ㄱ  
        case 42: return 1;  // ㄲ  
        case 44: return 2;  // ㄴ  
        case 47: return 3;  // ㄷ  
        case 48: return 5;  // ㄹ  
        case 56: return 6;  // ㅁ  
        case 57: return 7;  // ㅂ  
        case 59: return 9;  // ㅅ  
        case 60: return 10; // ㅆ  
        case 61: return 11; // ㅇ  
        case 62: return 12; // ㅈ  
        case 63: return 14; // ㅊ  
        case 64: return 15; // ㅋ  
        case 65: return 16; // ㅌ  
        case 66: return 17; // ㅍ  
        case 67: return 18; // ㅎ  
        }  
        return -1;  
    }  
      
    // 종성으로 변환  
    private int toFinal(int nKeyCode)  
    {  
        switch(nKeyCode)  
        {  
        case 0: return 41;  // ㄱ  
        case 1: return 42;  // ㄲ  
        case 2: return 44;  // ㄴ  
        case 3: return 47;  // ㄷ
        case 5: return 48;  // ㄹ  
        case 6: return 56;  // ㅁ  
        case 7: return 57;  // ㅂ  
        case 9: return 59;  // ㅅ  
        case 10: return 60; // ㅆ  
        case 11: return 61; // ㅇ  
        case 12: return 62; // ㅈ  
        case 14: return 63; // ㅊ  
        case 15: return 64; // ㅋ  
        case 16: return 65; // ㅌ  
        case 17: return 66; // ㅍ  
        case 18: return 67; // ㅎ  
        }  
        return -1;  
    }  
    
    // 
    private void decomposeWord(char ch) {
    	
    	if (ch < 0x314F) {	// ㅏ // 0x314F or 12623
    		prevWordClear();
    		return;
    	}
    	int uniValue = ch - BASE_CODE;
  
    	int jong = uniValue % 28;
    	int jung = (( uniValue - jong ) / 28 ) % 21;
    	int cho = (( uniValue - jong ) / 28 ) / 21;
    	
    	prevWordPhonemes[0] = cho;
    	prevWordPhonemes[1] = jung + 19;
    	prevWordPhonemes[2] = jong + 40;
    }
      
    // 자음분해  
    private void decomposeConsonant()  
    {  
        int i = 0;  
        if(curWordPhonemes[3] > 40 || curWordPhonemes[4] > 40)  
        {  
            do  
            {  
                if(DIVIDE_JONG_CONSON[i][2] == curWordPhonemes[2])  
                {  
                    curWordPhonemes[3] = DIVIDE_JONG_CONSON[i][0];  
                    curWordPhonemes[4] = DIVIDE_JONG_CONSON[i][1];  
      
                    completeWord = combineHangle(3);  
                    curWordPhonemes[0]   = toInitial(curWordPhonemes[4]);  
                    return;  
                }  
            }  
            while(++i< 11);  
        }  
      
        completeWord = combineHangle(1);  
        curWordPhonemes[0]   = toInitial(curWordPhonemes[2]);  
    }
    
    // 자음분해하여 prevWord에 저장  
    private boolean decomposeConsonant2()  
    {  
        int i = 0;  
        if(prevWordPhonemes[2] > 40)  
        {  
            do  
            {  
                if(DIVIDE_JONG_CONSON[i][2] == prevWordPhonemes[2])  
                {  
                	prevWordPhonemes[3] = DIVIDE_JONG_CONSON[i][0];  
                	prevWordPhonemes[4] = DIVIDE_JONG_CONSON[i][1];  
                    return true;  
                }  
            }  
            while(++i< 11);  
        }  
        return false;
    }
    
    /*// 사용하지 않음
    // 초성합성  
    private boolean mixInitial(int nKeyCode)  
    {  
        if(nKeyCode >= 19)  
            return false;  
      
        int i = 0;  
        do  
        {  
            if(MIXED_CHO_CONSON[i][0] == curWordPhonemes[0] && MIXED_CHO_CONSON[i][1] == nKeyCode)  
            {  
            	curWordPhonemes[0] = MIXED_CHO_CONSON[i][2];  
                return true;  
            }  
        }  
        while(++i < 14);  
      
        return false;  
    }*/
      
    
    // 종성합성  
    private boolean mixFinal(int nKeyCode)  
    {  
        if(nKeyCode <= 40) return false;  
      
        int i = 0;  
        do  
        {  
            if(MIXED_JONG_CONSON[i][0] == curWordPhonemes[2] && MIXED_JONG_CONSON[i][1] == nKeyCode)  
            {  
                curWordPhonemes[3] = curWordPhonemes[2];  
                curWordPhonemes[4] = nKeyCode;  
                curWordPhonemes[2] = MIXED_JONG_CONSON[i][2];  
      
                return true;  
            }  
        }  
        while(++i < 10);  
      
        return false;  
    } 
    
    // 종성합성  
    private boolean mixFinal2(int nKeyCode)  
    {  
        if(nKeyCode <= 40) return false;  
      
        int i = 0;  
        do  
        {  
            if(MIXED_JONG_CONSON[i][0] == prevWordPhonemes[2] && MIXED_JONG_CONSON[i][1] == nKeyCode)  
            {  
            	prevWordPhonemes[3] = prevWordPhonemes[2];  
            	prevWordPhonemes[4] = nKeyCode;  
            	prevWordPhonemes[2] = MIXED_JONG_CONSON[i][2];  
      
                return true;  
            }  
        }  
        while(++i < 10);  
      
        return false;  
    }  
      
    // 모음합성  
    private int mixVowel(int currentCode, int inputCode)  
    {  
        int i = 0;  
        do  
        {  
            if(MIXED_VOWEL[i][0] == currentCode && MIXED_VOWEL[i][1] == inputCode)  
            {  
                return MIXED_VOWEL[i][2];  
            }  
        }  
        while(++i< 13);  
      
        return 1000;  
    }  
      
    // 한글조합  
    private char combineHangle(int cho, int jung, int jong)  
    {  
        // 초성 * 21 * 28 + (중 성 - 19) * 28 + (종 성 - 40) + BASE_CODE;  
        return (char)(BASE_CODE - 572 + jong + cho * 588 + jung * 28);  
    }  
      
    private char combineHangle(int status)  
    {  
        switch(status)  
        {  
        //초 성 + 중성  
        case 1: return combineHangle(curWordPhonemes[0], curWordPhonemes[1], 40);  
      
        //초성 + 중성 + 종성  
        case 2: return combineHangle(curWordPhonemes[0], curWordPhonemes[1], curWordPhonemes[2]);  
          
        //초 성 + 중성 + 곁자음01  
        case 3: return combineHangle(curWordPhonemes[0], curWordPhonemes[1], curWordPhonemes[3]);  
        }  
        return ' ';  
    }  
      
    private void combineIngWord(int status)  
    {  
        switch(wordStatus)  
        {  
        case HS_FIRST:  
        case HS_FIRST_V:  
        case HS_FIRST_C:  
            ingWord = SOUND_TABLE[curWordPhonemes[0]];  
            break;  
      
        case HS_MIDDLE_STATE:  
        case HS_END:  
            ingWord = combineHangle(1);  
            break;  
      
        case HS_END_STATE:  
        case HS_END_EXCEPTION:  
            ingWord = combineHangle(2);  
            break;  
        }  
    }  
      
    private int downGradeIngWordStatus(char word)  
    {  
        int iWord = word;  
      
        //초성만 있는 경우  
        //  
        if(iWord < LIMIT_MIN || iWord > LIMIT_MAX)  
        {  
            ingWord = 0;
            return HS_FIRST;  
        }  
      
        //문자코드 체계  
        //iWord = firstWord * (21*28)  
        //      + middleWord * 28  
        //      + lastWord * 27  
        //      + BASE_CODE;  
        //  
        int totalWord   = iWord - BASE_CODE;  
        int iFirstWord  = totalWord / 28 / 21;  //초성  
        int iMiddleWord = totalWord / 28 % 21;  //중성  
        int iLastWord   = totalWord % 28;       //종성  
      
        curWordPhonemes[0] = iFirstWord; //초성저장  
      
        if(iLastWord == 0)  //종성이 없는 경우  
        {  
            ingWord = SOUND_TABLE[curWordPhonemes[0]];  
            return HS_FIRST_V;  
        }  
      
        curWordPhonemes[1] = iMiddleWord + 19; //중성저장  
      
        iLastWord += 40;  
      
        for(int i = 0; i < 11; i++)  
        {  
            if(iLastWord == DIVIDE_JONG_CONSON[i][2])  
            {  
                ingWord = combineHangle(3);  
                curWordPhonemes[2] = DIVIDE_JONG_CONSON[i][0]; // 종성저장  
                return HS_END_EXCEPTION;  
            }  
        }  
      
        ingWord = combineHangle(1);  
      
        return HS_MIDDLE_STATE;  
    }    
}
